package JointFeldmanProtocol;

import FeldmanVerifiableSecretSharing.ShamirSecretSharing.Polynomial;
import FeldmanVerifiableSecretSharing.VerifiableSecretSharing;
import com.google.protobuf.ByteString;
import com.google.protobuf.InvalidProtocolBufferException;
import org.bouncycastle.util.Arrays;
import meerkat.protobuf.DKGMessages.*;

import java.math.BigInteger;
import java.util.HashSet;
import java.util.Random;
import java.util.Set;
import java.util.concurrent.Callable;

/**
 * Created by Tzlil on 2/5/2016.
 *
 * an implementation of a version of Pedersen's distributed key generation protocol
 */
public class DKG extends VerifiableSecretSharing implements Runnable{

    private final int id;
    private final Network.User user; // send and receive messages throw network

    private final BigInteger[] ys; // container for y values that
    private final Polynomial.Point[] shares; // shares[i] equivalent to Si,id in terms of the protocol
    private final BigInteger[][] commitmentsArray; // commitmentsArray[i] equivalent to Ai in terms of the protocol
    private final ComplainState[][] complainStates; // complainStates[i][j] == state of Pj's complaint against Pi

    private final Set<Integer> QUAL; // set of all non-disqualified parties
    private final BigInteger[] commitments; // public verification values
    private Polynomial.Point share; // final share of the secrete
    private BigInteger y; // final public value


    public DKG(int t, int n, BigInteger x, Random random, BigInteger p, BigInteger q, BigInteger g,Network network) {
        super(t, n, x, random, p, q, g);

        this.commitmentsArray = new BigInteger[n][t + 1];
        this.shares = new Polynomial.Point[n];
        this.user = network.connect(new Handler());
        this.id = user.getID();
        this.complainStates = new ComplainState[n][n];
        this.QUAL = new HashSet<Integer>();
        this.ys = new BigInteger[n];
        this.commitments = new BigInteger[t + 1];

        for (int i = 0; i < n; i ++){
            for (int j = 0 ; j < n ; j ++)
                complainStates[i][j] = ComplainState.Non;
        }

    }

    /**
     * use for simulate real distributed protocol
     */
    @Override
    public void run() {
        user.getReceiverThread().start();
        stage1();
        stage2();
        stage3();
        stage4();
        user.getReceiverThread().interrupt();
    }

    /**
     * stage1 according to the protocol
     *  1. Pi broadcasts Aik for k = 0,...,t.
     *  2. Pi computes the shares Sij for j = 1,...,n and sends Sij secretly to Pj.
     */
    private void stage1(){
        // for avoiding edge cases, sets commitmentsArray[id - 1]
        BigInteger[] commitments = super.getCommitments();
        System.arraycopy(commitments, 0, commitmentsArray[id - 1], 0, commitmentsArray[id - 1].length);

        // broadcasts commitments
        CommitmentMessage commitment;
        for (int k = 0; k <= t ; k ++){
            commitment = CommitmentMessage.newBuilder()
                            .setK(k)
                            .setCommitment(ByteString.copyFrom(commitmentsArray[id - 1][k].toByteArray()))
                            .build();
            user.broadcast(Mail.Type.COMMITMENT,commitment);
        }

        // computes and sends shares
        SecretMessage secret;
        for (int j = 1; j <= n ; j++ ){
            if(j != id){
                secret = SecretMessage.newBuilder()
                            .setSecret(getShare(j).asMessage())
                            .build();
                user.send(j, Mail.Type.SECRET,secret);
            }
            else{
                shares[id - 1] = super.getShare(id);
            }
        }
        while (!isStage1Complete()){
            try {
                Thread.sleep(300);
            } catch (InterruptedException e) {
                // do nothing
            }
        }
    }

    /**
     * @return true iff all shares and commitments were received
     */
    private boolean isStage1Complete(){
        for (int i = 1 ; i <= n ; i++){
            if(shares[i - 1] == null)
                return false;
        }

        for (int i = 0; i < commitmentsArray.length; i++){
            for (int j = 0; j < commitmentsArray[i].length; j++){
                if(commitmentsArray[i][j] == null)
                    return false;
            }
        }
        return true;
    }

    /**
     * @param secret
     * @param i
     * @return g ^ Sij == verify(j,Ai,zpstar) (mod p)
     */
    private boolean isValidSecret(Polynomial.Point secret, int i){
        int j = secret.x.intValue();
        return zpstar.multiply(g,secret.y).equals(verify(j,commitmentsArray[i - 1],zpstar));
    }

    /**
     * stage2 according to the protocol
     *  Pj verifies all the shares he received (using isValidSecret)
     *  if check fails for an index i, Pj broadcasts a complaint against Pi.
     *  Pj broadcasts yj value at the end of this stage
     */
    private void stage2(){
        ys[id - 1] = super.getY();
        ComplaintMessage complaint;
        for (int i = 1; i <= n ; i++ ){
            if(id != i && !isValidSecret(shares[i - 1],i)) {
                //message = new Message(Type.Complaint, j)
                complaint = ComplaintMessage.newBuilder()
                                .setId(i)
                                .build();
                user.broadcast(Mail.Type.COMPLAINT,complaint);
                complainStates[i - 1][id - 1] = ComplainState.Waiting;
            }
        }
        //broadcast y after all complaints
        YMessage yMessage = YMessage.newBuilder()
                .setY(ByteString.copyFrom(super.getY().toByteArray()))
                .build();
        user.broadcast(Mail.Type.Y,yMessage);

        while (!isStage2Complete()){
            try {
                Thread.sleep(300);
            } catch (InterruptedException e) {
                // do nothing
            }
        }
    }

    /**
     * @return true iff all yi received for i = 1,...,n .
     */
    private boolean isStage2Complete() {
        for (int j = 1; j <= n ; j++) {
            if (j != id && ys[j - 1] == null)
                return false;
        }
        return true;
    }


    /**
     * stage3 according to the protocol
     *  1. if more than t players complain against a player Pi he is disqualified.
     *  2. Pi broadcasts the share Sij for each complaining player Pj.
     *  3. if any of the revealed shares fails the verification test, player Pi is disqualified.
     *  4. set QUAL to be the set of non-disqualified players.
     */
    private void stage3(){

        // broadcasts Sij for each complaint against Pid
        for (int j = 1 ; j <= complainStates[id - 1].length;j++) {
            switch (complainStates[id - 1][j - 1]) {
                case Waiting:
                    user.broadcast(Mail.Type.SECRET, SecretMessage.newBuilder()
                            .setSecret(getShare(j).asMessage())
                            .build());
                    complainStates[id - 1][j - 1] = ComplainState.NonDisqualified;
                    break;
                default:
                    break;
            }
        }

        // wait until there is no complaint waiting for answer
        for (int i = 0; i < complainStates.length;i++){
            for (int j = 0 ; j < complainStates[i].length;j++){
                while (complainStates[i][j].equals(ComplainState.Waiting)){
                    try {
                        Thread.sleep(300);
                    } catch (InterruptedException e) {
                        // do nothing
                    }
                }
            }
        }

        // add each non-disqualified player to QUAL
        boolean nonDisqualified;
        for (int i = 1; i <= complainStates.length;i++){
            nonDisqualified = true;
            for (int j = 1 ; j <= complainStates[i - 1].length;j++){
                nonDisqualified &= complainStates[i - 1][j - 1].equals(ComplainState.Disqualified);
            }
            if(nonDisqualified){
                QUAL.add(i);
            }
        }
    }

    /**
     * stage4 according to the protocol
     *  1. public value y is computed as y = multiplication of yi mod p for i in QUAL
     *  2. public verification values are computed as Ak = multiplication of Aik mod p for i in QUAL for k = 0,...,t
     *  3. Pj sets is share of the secret as xj = sum of Sij mod q for i in QUAL
     */
    private void stage4(){
        this.y = zpstar.zero();
        for (int i : QUAL) {
            this.y = zpstar.add(this.y , ys[i - 1]);
        }
        BigInteger commitment;
        for (int k = 0; k <= t ; k++){
            commitment = zpstar.zero();
            for (int i : QUAL) {
                commitment = zpstar.add(commitment,commitmentsArray[i - 1][k]);
            }
            commitments[k] = commitment;
        }

        BigInteger xj = BigInteger.ZERO;
        for (int i : QUAL) {
            xj = xj.add(shares[i - 1].y);
        }
        this.share = new Polynomial.Point(BigInteger.valueOf(id) , xj.mod(q));
    }

    @Override
    public BigInteger getY() {
        return y;
    }

    @Override
    public BigInteger[] getCommitments() {
        return Arrays.clone(commitments);
    }




    private enum ComplainState{
        Non, Waiting,Disqualified,NonDisqualified
    }

    private class Handler implements Network.MailHandler {

        private Handler() {}

        void handelSecretMessage(Mail mail) throws InvalidProtocolBufferException {
            if(shares[mail.getSender() - 1] == null) {
                SecretMessage secretMessage = SecretMessage.parseFrom(mail.getMessage());
                Polynomial.Point secret = new Polynomial.Point(secretMessage.getSecret());
                if(mail.getIsPrivate()){
                    shares[mail.getSender() - 1] = secret;
                }else{
                    int i = mail.getSender();
                    int j = secret.x.intValue();
                    switch (complainStates[i - 1][j - 1]){
                        case Waiting:
                            if(isValidSecret(secret,i)){
                                complainStates[i - 1][j - 1] = ComplainState.NonDisqualified;
                            }else{
                                complainStates[i - 1][j - 1] = ComplainState.Disqualified;
                            }
                            break;
                        default:
                            break;
                    }
                }
            }
        }

        void handelCommitmentMessage(Mail mail) throws InvalidProtocolBufferException {
            if(!mail.getIsPrivate()) { //broadcast only
                CommitmentMessage commitmentMessage = CommitmentMessage.parseFrom(mail.getMessage());
                if (commitmentsArray[mail.getSender() - 1][commitmentMessage.getK()] == null) {
                    BigInteger commitment = new BigInteger(commitmentMessage.getCommitment().toByteArray());
                    commitmentsArray[mail.getSender() - 1][commitmentMessage.getK()] = commitment;
                }
            }
        }

        void handelYMessage(Mail mail) throws InvalidProtocolBufferException {
            if(!mail.getIsPrivate()) { //broadcast only
                if (ys[mail.getSender() - 1] == null) {
                    YMessage yMessage = YMessage.parseFrom(mail.getMessage());
                    BigInteger y = new BigInteger(yMessage.getY().toByteArray());
                    ys[mail.getSender() - 1] = y;
                }
            }
        }

        void handelComplaintMessage(Mail mail) throws InvalidProtocolBufferException {
            int id = user.getID();
            if(!mail.getIsPrivate()) { //broadcast only
                ComplaintMessage complaintMessage = ComplaintMessage.parseFrom(mail.getMessage());
                int i = complaintMessage.getId();
                int j = mail.getSender();
                switch (complainStates[i - 1][j - 1]){
                    case Non:
                        complainStates[i - 1][j - 1] = ComplainState.Waiting;
                        break;
                    default:
                        break;
                }
            }
        }

        @Override
        public void handel(Mail mail) throws InvalidProtocolBufferException {
           switch (mail.getType()){
               case SECRET:
                   handelSecretMessage(mail);
                   break;
               case COMMITMENT:
                   handelCommitmentMessage(mail);
                   break;
               case Y:
                   handelYMessage(mail);
                   break;
               case COMPLAINT:
                   handelComplaintMessage(mail);
                   break;
               default:
                   break;
           }
        }
    }
}
